# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
#
from __future__ import print_function, absolute_import
import os
import subprocess
import json
from copy import deepcopy

class Scheduler(object):
    """
    Generic scheduler class

    Initialize the class, by default we look for a config file in `~/.yambopy/config.json`

    This file has the following information:

        .. code-block:: javascript

            {
                "default": "oar <or> bash <or> pbs",
                "oar <or> bash <or> pbs":
                {
                    "mpirun": "<mpirun command>",
                    "modules": {"<module tag>":"<module name>"},
                    "pre_run": ["line1 to execute before the run",
                                "line2 to execute before the run"],
                    "pos_run": "file:<file_in_config_folder>",
                    "<tag>":"<value>"
                }
            }

    The "default" tag chooses the scheduler to use when no scheduler is
    specified by the user.

    "<tag>" and "<value>" in are additional variables that are handled
    differently for each scheduler. They are stored as arguments in the variable kwargs.
    """

    #TODO: Add a way to get the home directory
    _config_path     = os.path.expanduser("~") + "/.yambopy"
    _config_filename = "%s/config.json"%_config_path

    def __init__(self, name=None, nodes=None, cores=None, cpus_per_task=None, walltime="1:00:00", **kwargs ):
        self.name = name
        self.nodes = nodes
        self.cores = cores
        self.cpus_per_task = cpus_per_task
        self.walltime = walltime
        self.kwargs = kwargs

        self.pre_run = self.get_arg("pre_run",[])
        self.pos_run = self.get_arg("pos_run",[])

        self.modules_dict = self.get_arg("modules_dict",{})
        self.modules_list = []
        for mod in self.get_arg("modules_list",[]):
            self.add_module(mod)
        self.arguments = []
        self.commands  = []

        self.initialize()


    @classmethod    
    def factory(cls,scheduler=None,cores=None,nodes=None,walltime="1:00:00",**kwargs):
        """
        Initialize a scheduler instance.

        Default arguments:
        cores    - Number of cores to use
        nodes    - Number of nodes to use
        walltime - Walltime limit
        **kwargs - Additional tags specific for each scheduler
        """
        from .oar import Oar
        from .pbs import Pbs
        from .bash import Bash
        from .slurm import Slurm

        schedulers = { "oar":   Oar,
                       "pbs":   Pbs,
                       "slurm": Slurm,
                       "bash":  Bash }

        #load configurations file
        config = cls.load_config()

        #set from scheduler
        schedulername = scheduler
        if not scheduler: schedulername = config.get('default',None)

        if schedulername is None:
            raise ValueError('scheduler not specified and "default" not set in configuration file %s.'%self._config_filename)

        #load the configurations
        schedulerconfig = config.get(schedulername,{})

        #determine the scheduler type
        if "type" in schedulerconfig:
            schedulertype = schedulerconfig["type"]
        elif "oar"   in schedulername: schedulertype = "oar"
        elif "pbs"   in schedulername: schedulertype = "pbs"
        elif "bash"  in schedulername: schedulertype = "bash"
        elif "slurm" in schedulername: schedulertype = "slurm"
        else:
            raise ValueError("Could not determine the scheduler type. "
                             "Please specify it in the name of the scheduler or using the 'type' tag.")

        #sanity check
        if schedulername not in list(schedulers.keys()):
            raise ValueError("Scheduler name %s is invalid"%schedulername)

        #check type from outside class
        cls.schedulertype = schedulertype

        if "nodes" in schedulerconfig and nodes is None:
            nodes = int(schedulerconfig["nodes"])
            del schedulerconfig["nodes"]
        if "cores" in schedulerconfig and cores is None:
            cores = int(schedulerconfig["cores"])
            del schedulerconfig["cores"]

        walltime = schedulerconfig.pop("walltime",walltime)

        #add scheduler config arguments
        kwargs.update(schedulerconfig)

        #create an instance of the scheduler to use
        return schedulers[schedulertype](cores=cores,nodes=nodes,walltime=walltime,**kwargs)

    @staticmethod
    def load_config():
        """
        load the configuration file and check its sanity
        """
        try:
            with open(Scheduler._config_filename) as f:
                config = json.load(f)
        except IOError:
            config = {  "default":"bash",
                        "bash":
                        {
                            "mpirun": "mpirun",
                            "modules": "None",
                            "pre_run": [],
                            "pos_run": []
                        }
                     }

        #TODO: put sanity checks here
        schedulername = "bash"
        if 'default' in config:
            schedulername = config['default']

        return config

    def initialize(self):
        raise NotImplementedError("Initialize not implemented in this class!")

    def clean(self):
        """
        clean the command list
        """
        self.commands = []
        self.modules_list = []

    def get_vardict(self):
        """
        initialize the vardict from the default values and check in the arguments if
        a new assignment of the variable was specified.
        For example to assing the cores to ppn the user can set in the configurations file the
        "var_cores":"ppn" or to the arguments of the scheduler.factory() function as var_cores="ppn".
        """
        #load default vardict
        self.vardict = deepcopy(self._vardict)
        cores = self.get_arg('var_cores')
        if cores: self.vardict['cores'] = cores
        nodes = self.get_arg('var_nodes')
        if nodes: self.vardict['nodes'] = nodes
    
    def copy(self):
        """
        return a copy of this instance
        """
        return deepcopy(self)

    def write(self,filename):
        """
        write the bash file to a file in the disk
        """
        if not filename: filename = self.name
        with open(filename,"w") as f:
            f.write(str(self))

    def get_arg(self,argument,default=None):
        """
        get an argument from the list of optional variables kwargs

        return:
        None - if the variable does not exist
        value - if the variable exists return the value of the variable
        """
        if argument in self.kwargs:
            arg = self.kwargs[argument]
            if arg == "true":
                return True
            elif arg == "false":
                return False
            elif type(arg) in [str,str] and "file:" in arg:
                #load the text from a textfile
                filename = arg.split(":")[-1]
                f = open("%s/%s"%(self._config_path,filename),'r')
                return f.read().strip().split('\n')
                f.close()
            else:
                return self.kwargs[argument]
        else:
            return default

    def add_command(self,cmd):
        """
        add commands to be run by the scheduler
        """
        self.commands.append(cmd)

    def add_arguments(self,arguments):
        """
        add commands to be run by the scheduler
        """
        self.arguments.append(arguments)

    def add_mpirun_command(self,cmd):
        """
        add commands to be run by the scheduler
        """
        threads = 1
        if self.cores: threads*=self.cores
        if self.nodes: threads*=self.nodes
        mpirun = self.get_arg("mpirun","mpirun")
        np = self.get_arg("np","-np")
        self.add_command("%s %s %d %s"%(mpirun,np,threads,cmd))

    def add_module(self,mod):
        """
        Add module to be loaded by the scheduler
        
        Arguments:
            mod: if the mod string exists in the "modules_dict" dictionary get the key value
                 if it does not exist simply add the string assuming the user knows what he is doing
        """
        if self.modules_dict:
          if mod in self.modules_dict:
            self.modules_list.append(self.modules_dict[mod])
            return
        self.modules_list.append(mod)

    def run(self):
        """
        run the command in the bash
        """
        raise NotImplementedError('Run not implemented')

    def set_posrun(self,posrun):
        self.pos_run = posrun

    def set_prerun(self,prerun):
        self.pre_run = prerun
       
    @property
    def modulelist(self):
        return ["module load %s"%mod for mod in self.modules_list]

    def get_commands(self):
        """
        get commands to execute
        """
        lines = []; app = lines.append
        if self.pre_run:
            app("\n#pre_run")
            lines += self.pre_run
        if self.modules_list:
            app("\n#modules")
            lines += self.modulelist
        app("\n#commands")
        lines += self.commands
        if self.pos_run:
            app("\n#pos_run")
            lines += self.pos_run
        return "\n".join(lines)

if __name__ == "__main__":
    #after
    s = Scheduler(nodes=1,cores=4)
    s.add_command('abinit < run.sh')
    print(s)



##
## Authors: MN (FP adapted)
##
import numpy as np
import os
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.bse.exciton_matrix_elements import exciton_X_matelem
from yambopy.bse.rotate_excitonwf import rotate_exc_wf
from tqdm import tqdm

def exciton_phonon_matelem(latdb,elphdb,wfdb,Qrange=[0,1],BSE_dir='bse',BSE_Lin_dir=None,
                           nexc_in=-1,nexc_out=-1,dmat_mode='run',save_files=True,exph_file='Ex-ph.npy',overwrite=False):
    """
    This function calculates the exciton-phonon matrix elements

    - Q is the exciton momentum
    - q is the phonon momentum
    - exc_in represent the "initial" exciton states in the scattering process (at mom. Q)
    - exc_out represents the "final" exciton states in the scattering process (at mom. Q+q)

    Parameters
    ----------
    latdb : YamboLatticeDB
        The YamboLatticeDB object which contains the lattice information.
    elphdb : LetzElphElectronPhononDB
        The LetzElphElectronPhononDB object which contains the electron-phonon matrix
        elements.
    wfdb : YamboWFDB
        The YamboWFDB object which contains the wavefunction information.
    BSE_dir : str, optional
        The name of the folder which contains the BSE calculation. Default is 'bse'.
    BSE_Lin_dir : str, optional
        The name of the folder which contains the BSE Q=0 calculation (for optical spectra). Default is BSE_dir.
    Qrange : int list, optional
        Exciton Qpoint index range [iQ_initial, iQ_final] (python counting). Default is [0,0] (Gamma point only).
        Note that the indexing is in full BZ and not in iBZ. See wfc.kBZ to see the kpoints in full BZ
    nexc_in : int, optional
        Number of excitonic states included in Lin calculation. Default is -1 (all).
    nexc_out : int, optional
        Number of excitonic states included in Lout calculation. Default is -1 (all).
    dmat_mode : str, optional
        If 'save', print dmats on .npy file for faster recalculation. If 'load', load from .npy file. Else, calculate Dmats at runtime.
    save_files : bool, optional
        If True, the matrix elements will be saved in .npy file `exph_file`. Default is True.
    overwrite : bool, optional
        If False and `exph_file` is found, the matrix elements will be loaded from file. Default is False.
    """

    # Check if we just need to load
    if os.path.exists(exph_file) and overwrite==False:
        print(f'Loading EXCPH matrices from {exph_file}...')
        exph_mat_loaded = np.load(exph_file)
        return exph_mat_loaded

    # Load exc dbs
    exdbs = []
    for ik in range(wfdb.nkpoints):
        filename = 'ndb.BS_diago_Q%d' % (ik+1)
        excdb = YamboExcitonDB.from_db_file(latdb,filename=filename,folder=BSE_dir,\
                                            Load_WF=True, neigs=nexc_out)
        exdbs.append(excdb)
        #
        # NM : Add a sanity check to avoid a disasterous consequence
        # if the user gives wrong bse band indices.
        min_bnd_bse = np.min(excdb.unique_vbands)
        max_bnd_bse = np.max(excdb.unique_cbands)+1
        assert (wfdb.min_bnd == min_bnd_bse) and ((wfdb.min_bnd + wfdb.nbands) == max_bnd_bse), \
            print("Error: BSE bands mismatch. Given bands range : [%d, %d]. " %(
                wfdb.min_bnd,wfdb.min_bnd + wfdb.nbands) +
                "Bse band range found (expected) : [%d %d]" %( min_bnd_bse,max_bnd_bse))

    # get D matrices
    Dmats = save_or_load_dmat(wfdb,mode=dmat_mode,dmat_file='Dmats.npy')

    # Calculation
    print('Calculating EXCPH matrix elements...')
    exph_mat = []
    for iQ in tqdm(range(Qrange[0],Qrange[1])):
        Q_in = wfdb.kBZ[iQ]
        exph_mat.append( exciton_phonon_matelem_iQ(elphdb,wfdb,exdbs,Dmats,\
                                                   BSE_Lin_dir=BSE_Lin_dir,Q_in=Q_in,neigs=nexc_in) )
    # IO
    if len(exph_mat)<2: exph_mat = exph_mat[0] # single Q-point calculation (suppress axis)
    else:               exph_mat = np.array(exph_mat) #[nQ,nq,nmodes,nexc_in (Qexc),nexc_out (Qexc+q)]
    
    if save_files: 
        if exph_file[-4:]!='.npy': exph_file = exph_file+'.npy'
        print(f'Excph coupling file saved to {exph_file}')
        np.save(exph_file,exph_mat)
    
    return exph_mat

def exciton_phonon_matelem_iQ(elphdb,wfdb,exdbs,Dmats,BSE_Lin_dir=None,
                              Q_in=np.zeros(3),neigs=-1,dmat_mode='run'): 
    """
    This function calculates the exciton-phonon matrix element per Q 

    - Q is the exciton momentum
    - q is the phonon momentum
    - exc_in represent the "initial" exciton states in the scattering process (at mom. Q)
    - exc_out represents the "final" exciton states in the scattering process (at mom. Q+q)

    Parameters
    ----------
    elphdb : LetzElphElectronPhononDB
        The LetzElphElectronPhononDB object which contains the electron-phonon matrix
        elements.
    wfdb : YamboWFDB
        The YamboWFDB object which contains the wavefunction information.
    exdbs : YamboExcitonDB list
        List of Q+q YamboExcitonDB objects containing the BSE calculation
    BSE_Lin_dir : str, optional
        The name of the folder which contains the BSE q=0 calculation (for optical spectra). Default is exdbs[Q].
    Q_in : np.ndarray, optional
        Excitonic momentum in reduced units. Default np.array([0.0,0.0,0.0]) 
    neigs : int, optional
        Number of excitonic states included in calculation. Default is -1 (all).
    """
    latdb = wfdb.ydb
    # Determine Lkind(in)
    Ak = rotate_Akcv_Q(wfdb, exdbs, Q_in, Dmats, neigs=neigs, folder=BSE_Lin_dir)
    # Compute ex-ph
    exph_mat = []
    bse_bnds_range = [wfdb.min_bnd,wfdb.min_bnd + wfdb.nbands]
    for iq in range(elphdb.nq):
        ph_eig, elph_mat = elphdb.read_iq(iq,bands_range=bse_bnds_range,convention='standard')
        elph_mat = elph_mat.transpose(1,0,2,4,3)
        #
        Akq = rotate_Akcv_Q(wfdb, exdbs, Q_in + elphdb.qpoints[iq], Dmats) # q+Q
        tmp_exph = exciton_X_matelem(Q_in, elphdb.qpoints[iq], \
                                     Akq, Ak, elph_mat, wfdb.kBZ, \
                                     contribution='b', diagonal_only=False, ktree=wfdb.ktree)
        exph_mat.append(tmp_exph)

    ## 0.5 for Ry to Ha
    exph_mat = 0.5 * np.array(exph_mat).transpose(0,1,3,2) #[nq,nmodes,nexc_in (Qexc),nexc_out (Qexc+q)]

    return exph_mat

def save_or_load_dmat(wfdb, mode='run', dmat_file='Dmats.npy'):
    """
    Save or load Dmats to/from .npy file `dmat_file` for faster recalculation.

     If mode=='save', print dmats on .npy file for faster recalculation. 
     If mode=='load', load from .npy file. 
     Else, calculate Dmats at runtime.
    """
    if dmat_file[-4:]!='.npy': dmat_file = dmat_file+'.npy'
    if mode=='save':
        print('Saving D matrices...')
        Dmats = wfdb.Dmat()
        np.save(dmat_file,Dmats)
        return Dmats
    elif mode=='load': 
        print('Loading D matrices...')
        if not os.path.exists(dmat_file):
            raise FileNotFoundError(f"Cannot load '{dmat_file}' - file does not exist.")
        Dmats_loaded = np.load(dmat_file)
        return Dmats_loaded
    else:
        return wfdb.Dmat()


def rotate_Akcv_Q(wfdb, exdbs, Qpt, Dmats, neigs=-1, folder=None):
    '''
    Qpt reduced coordinates in BZ or whatever

    wfdb : wavefunction object
    exdbs : previously read list of exciton objects for Lout (used if folder is None)
    Qpt: reduced coordinates in BZ or whatever
    Dmats: precalculated Dmat
    neigs: number of states (used if folder is not None, otherwise it's nexc_out)
    folder: where to load the Lin exciton states (if needed). 
            can be used also to reload the same in case we want nexc_in/=nexc_out for same L
    '''
    latdb = wfdb.ydb
    idx_BZQ = wfdb.kptBZidx(Qpt)
    iQ_isymm = latdb.symmetry_indexes[idx_BZQ]
    iQ_iBZ = latdb.kpoints_indexes[idx_BZQ]
    trev  = (iQ_isymm >= len(latdb.sym_car) / (1 + int(np.rint(latdb.time_rev))))
    symm_mat_red = latdb.lat@latdb.sym_car[iQ_isymm]@np.linalg.inv(latdb.lat)
    exe_iQIBZ = wfdb.kpts_iBZ[iQ_iBZ]
    #
    if folder is not None:
        if neigs==-1:
            neigs = len(exdbs[0].eigenvalues) # Set neigs equal to nexc_out
        filename = 'ndb.BS_diago_Q%d' % (iQ_iBZ+1)
        excdbin = YamboExcitonDB.from_db_file(latdb,filename=filename,folder=folder,\
                                              Load_WF=True, neigs=neigs)
        AQibz = excdbin.get_Akcv()
    else : AQibz = exdbs[iQ_iBZ].get_Akcv()
    #
    AQ_rot = rotate_exc_wf(AQibz,symm_mat_red,wfdb.kBZ,exe_iQIBZ,Dmats[iQ_isymm],trev,wfdb.ktree)
    
    return AQ_rot



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
import os
import json
import re
from subprocess import Popen, PIPE
from yambopy.env import yambopyenv
from yambopy.tools.duck import isstring


def issave(path):
    """ Check if yambo SAVE folder is present either as directory or as symlink """
    if os.path.isdir(path) or os.path.islink(path): return True
    else: return False

class YamboIn(object):
    """
    Class to read, write, create and manipulate yambo input files with python.

    Examples of use:

    Initialize an empty input file:

        .. code-block:: python

            y = YamboIn(filename='somefile.in')
            print y

    Call yambo to initialize the input file with variables according to the runlevel,
    parse the input file and store the variables:

        .. code-block:: python

            y = YamboIn('yambo -o c',folder='ip')
            print y

    If the argument ``args`` was used then the filename should be left as ``yambo.in`` because that's the default input filename that yambo will create.

    Call ypp to initialize the input file:

        .. code-block:: python

            y = YamboIn('yyp -e w'args=,filename='ypp.in')
            print y

    **Arguments:**

        ``args``:     if specified yambopy will run yambo, read the generated input file and initialize the class with those variables.

        ``folder``:   the folder where the SAVE directory is located

        ``vim``:      if yambo is compiled using vim as an editor this variable should be set as True because then `yambopy` will close vim.
        In newer versions an argument for yambo '-Q' tells it to not call vim

        ``filename``: the name of the input file to be read

    """
    #Regular expressions
    _variaexp   = '([A-Za-z\_0-9]+(?:\_[A-Za-z]+)?)' #variables names
    _numexp     = '([+-]?\d+(?:\.\d+)?(?:[eE][-+]?\d+)?)' #number
    _spacexp    = '(?:[ \t]+)?' #space
    _stringexp  = '["\']([a-zA-Z0-9_ ]+?)["\']' #string
    _arrayexp   = '%'+_spacexp+_variaexp+'\s+(?:\#.+)?((?:(?:\s|\.|[+-]?\d)+?\|)+)\s+([a-zA-Z]+)?' #arrays
    _complexexp = '\('+_spacexp+_numexp+_spacexp+','+_spacexp+_numexp+_spacexp+'\)' #complex numbers
    _runexp     = '([a-zA-Z0-9_]+)' #runlevels
    # list of available runlevels to be stored in the arguments array
    _runlevels  = ['rim_cut','chi','em1s','bse','optics','bsk','bss','dipoles','ExcitonGkkp'
                   'em1d','gw0','HF_and_locXC','setup','ppa','cohsex','life',
                   'collisions','negf','el_ph_scatt','el_el_scatt','excitons','wavefunction','fixsyms',
                   'QPDBs', 'QPDB_merge','RealTime','RT_X','RToccDos','RToccBnd','RToccEner',
                   'RToccTime','RTlifeBnd','amplitude','bzgrids','Random_Grid','gkkp','el_ph_corr','WRbsWF',
                   'Select_energy','RTDBs','photolum','kpts_map','RTtime','RToccupations','RTfitbands']

    def __init__(self,args=''):
        """
        Initalize the class
        """
        self.yamboargs = args

        #the type of the variables is determined from the type of variable in this dictionary
        self.variables = {} #here we will store the values of the variables
        self.arguments = [] #here we will store the arguments

    @classmethod
    def from_file(cls,filename='yambo.in',folder='.'):
        """ Create an instance of YamboIn from an existing input file"""
        instance = cls()
        err = instance.read_file(os.path.join(folder,filename))
        if err: raise FileNotFoundError('Could not read the %s file'%os.path.join(folder,filename))
        return instance

    @classmethod
    def from_runlevel(cls,runlevel,executable=yambopyenv.YAMBO,folder='.',filename='yambo.in'):
        """
        Create an input file from the runlevel.
        Will execute yambo in the folder with the runlevel arguments,
        read the file and return an instance of this class
        """
        workdir = os.getcwd()
        if executable[3:]=='ypp': filename='ypp.in'

        #check if there exists a SAVE folder
        save_path = os.path.join(folder,'SAVE')
        if not issave(save_path): raise ValueError('SAVE folder not found in %s'%save_path)

        #run yambo
        os.chdir(folder)
        if os.path.isfile(filename): os.remove(filename)
        if '-Q' not in runlevel: runlevel += ' -Q'
        if filename is not ('yambo.in' or 'ypp.in'): runlevel += ' -F %s'%filename
        command = "%s %s"%(executable,runlevel)
        yambo = Popen(command, stdout=PIPE, stderr=PIPE, stdin=PIPE, shell=True)
        yambo.wait()
        os.chdir(workdir)

        instance = cls()
        err = instance.read_file(os.path.join(folder,filename))

        if err:
            lines = []; app = lines.append
            app('Yambo did not create the %s input file.'%filename)
            app('command: %s'%command)
            app('folder:  %s/'%folder)
            raise FileNotFoundError("\n".join(lines))

        #read input file
        return cls.from_file(filename=filename,folder=folder)

    @classmethod
    def from_dictionary(cls,yamboin_dict):
        """Return an instance of this class from a dictionary"""
        yamboin = cls()
        yamboin.set_fromdict(yamboin_dict)
        return yamboin

    def set_fromdict(self,yamboin_dict):
        """Write a python script to generate this input"""
        #monkey patch the input file
        if 'arguments' in yamboin_dict.keys() and 'variables' in yamboin_dict.keys():
            for var,value in yamboin_dict.items():
                setattr(self,var,value)
        else:
            for var,value in yamboin_dict.items():
                self[var] = value

    def set_fromargs(self,yamboin_args):
        """Write a python script to generate this input"""
        #monkey patch the input file
        self.arguments += yamboin_args

    def __getitem__(self,key):
        """ Get the value of a variable in the input file
        """
        return self.variables[key]

    def __setitem__(self,key,value):
        """ Set the value of a variable in the input file
        """
        if key == "QPbands":
            #read from current QPkrange
            kstart,kstop,bstart,bstop = self.variables["QPkrange"][0]
            #read bands from QPbands
            new_bstart,new_bstop = value
            #set new QPkrange
            key = "QPkrange"
            value = [kstart,kstop,new_bstart,new_bstop]
        #if the units are not specified, add them
        if isinstance(value,list):
            if not any( [isinstance(v,str) for v in value] ):
                value = [value,'']
        if isinstance(value,(int,float,complex)):
            value = [value,'']
        self.variables[key] = value

    def __delitem__(self,key):
        """ Remove a keyword from the dicitonary
        """
        del self.variables[key]

    def read_file(self,filename='yambo.in'):
        """ Read the variables from a file
        """
        if not os.path.isfile(filename): return 1

        with open(filename,"r") as yambofile:
            inputfile = self.read_string(yambofile.read())
        return 0

    def add_dict(self,variables):
        """
        Add a dictionary containing variables to the current inputfile
        """
        self.variables.update(variables)

    def read_string(self,inputfile):
        """
        Read the input variables from a string
        """
        var_real     = re.findall(self._variaexp + self._spacexp + '='+ self._spacexp +
                                  self._numexp + self._spacexp + '([A-Za-z]+)?',inputfile)
        var_string   = re.findall(self._variaexp + self._spacexp + '='+ self._spacexp + self._stringexp, inputfile)
        var_array    = re.findall(self._arrayexp,inputfile)
        var_complex  = re.findall(self._variaexp + self._spacexp + '='+ self._spacexp + self._complexexp + self._spacexp + '([A-Za-z]+)?', inputfile)
        var_runlevel = re.findall(self._runexp + self._spacexp, inputfile)

        def clean(a):
            """
            clean the variables according to the type of data
            """
            a = a.strip()
            if a.replace('.','',1).isdigit():
                if "." in a: return float(a)
                else:        return int(a)
            return a

        # Determination of the arguments
        for key in self._runlevels:
            if key in var_runlevel:
                self.arguments.append(key)

        #float variables
        for var in var_real:
            name, value, unit = var
            self[name] = [float(value),unit]

        #string variables
        for var in var_string:
            name, string = var
            self[name] = string

        #complex variables
        for var in var_complex:
            name, real, imag, unit = var
            self[name] = [complex(float(real),float(imag)),unit]

        #array variables
        for var in var_array:
            name, array, unit = var
            array = [clean(val) for val in array.split('|')[:-1]]
            self[name] = [array,unit]

        return {"arguments": self.arguments, "variables": self.variables}

    def optimize(self,conv,folder='.',variables=('all',),run=lambda x: None,ref_run=True):
        """
        Function to to make multiple runs of yambo to converge calculation parameters

        Arguments:

            A dictionary conv that has all the variables to be optimized
            A list fo the name of the variables in the dicitonary that are to be optimized
            A function run that takes as input the name of the inputfile (used to run yambo)
            A boolean ref_run that can disable the submitting of the reference run (see scripts/analyse_gw.py)

        """
        name_files = []

        #check which variables to optimize
        if 'all' in variables:
            variables = list(conv.keys())

        #save all the variables
        backup = {}
        for var in variables:
            backup[var] = self[var]

        #add units to all the variables (to be improved!)
        for key,value in list(conv.items()):
            if not isinstance(value[-1],str) and type(value[0]) == list:
                conv[key] = [value,'']

        #make a first run with all the first elements
        reference = {}
        for key,value in list(conv.items()):
            values, unit = value
            reference[key] = [values[0],unit]
            self[key] = [values[0],unit]
        #write the file and run
        if ref_run==True:
            self.write( "%s/reference.in"%(folder) )
            run('reference.in')
        else:
            print('Reference run disabled.')

        #converge one by one
        for key in [var for var in list(conv.keys()) if var in variables]:
            values, unit = conv[key]
            #put back the original values of the variables
            for var in variables:
                self[var] = reference[var]
            #change the other variables
            if isinstance(values[0],str):
                for string in values[1:]:
                    filename = "%s_%s"%(key,string)
                    self[key] = string
                    self.write( folder + filename )
                    run(filename+".in")
                continue
            if type(values[0])==float:
                for val in values[1:]:
                    filename = "%s_%12.8lf"%(key,val)
                    self[key] = [val,unit]
                    self.write( "%s/%s.in"%(folder,filename) )
                    run(filename+".in")
                continue
            if type(values[0])==int:
                for val in values[1:]:
                    filename = "%s_%05d"%(key,val)
                    self[key] = [val,unit]
                    self.write( "%s/%s.in"%(folder,filename) )
                    run(filename+".in")
                continue
            if type(values[0])==list:
                for array in values[1:]:
                    filename = "%s_%s"%(key,"_".join(map(str,array)))
                    self[key] = [array,unit]
                    self.write( "%s/%s.in"%(folder,filename) )
                    run(filename+".in")
                continue
            if type(values[0])==complex:
                for value in values[1:]:
                    filename = "%s_%lf_%lf"%(key,value.real,value.imag)
                    self[key] = [value,unit]
                    self.write( "%s/%s.in"%(folder,filename) )
                    run(filename+".in")
                continue
            raise ValueError( "unknown type for variable:", key )

        #put back the original values of the variables
        for var in variables:
            self[var] = backup[var]

        return name_files

    def copy(self):
        """Return a copy of this object"""
        import copy
        return copy.deepcopy(self)

    def write(self,filename='yambo.in', prefix=''):
        """
        Write a yambo input file
        """
        with open(filename,"w") as f:
           f.write(prefix)
           f.write(str(self))

    def set_q(self,q):
        """Change one of ['QpntsRXp','QpntsRXd','QpntsRXs'] variables to calculate only one q-point"""
        for var in ['QpntsRXp','QpntsRXd','QpntsRXs']:
            qpts = self.variables.get(var,None)
            if qpts is None: continue
            self.variables[var] = [[q,q],'']
            return 0
        raise ValueError('Could not find one of the following variables set in the input file: \'QpntsRXp\',\'QpntsRXd\',\'QpntsRXs\'')

    def pack(self,filename):
        """
        Pack all the data of this structure in a `.json` file
        """
        f = open(filename,'w')
        json.dump(f,[self.arguments,self.real,self.string,self.complex,self.array],indent=5)
        f.close()

    def __str__(self):
        """
        Returns the input file as a string
        """
        s  = ""

        #arguments
        s += "\n".join(self.arguments)+'\n'

        for key,value in list(self.variables.items()):
            if key == 'DrudeWXd':
                s+= 'DrudeWXd= '+value+"\n"
                continue
            if isstring(value):
                s+= "%s = %10s\n"%(key,"'%s'"%value)
                continue
            if isinstance(value[0],float):
                val, unit = value
                if val > 1e-6:
                    s+="%s = %lf %s\n"%(key,val,unit)
                else:
                    s+="%s = %lf %s\n"%(key,val,unit)
                continue
            if isinstance(value[0],int):
                val, unit = value
                s+="%s = %d %s\n"%(key,val,unit)
                continue
            if isinstance(value[0],list):
                array, unit = value
                if isinstance(array[0],list):
                    s+='%% %s\n'%key
                    for l in array:
                        s+="%s \n"%(" | ".join(map(str,l))+' | ')
                    s+='%s'%unit
                    s+='%\n'
                else:
                    s+="%% %s\n %s %s \n%%\n"%(key," | ".join(map(str,array))+' | ',unit)
                continue
            if isinstance(value[0],str):
                array = value
                s+="%% %s\n %s \n%%\n"%(key," | ".join(["'%s'"%x.replace("'","").replace("\"","") for x in array])+' | ')
                continue
            if isinstance(value[0],complex):
                c, unit = value
                s+="%s = (%lf,%lf) %s\n"%(key,c.real,c.imag,unit)
                continue
            raise ValueError( "Unknown type %s for variable: %s" %( type(value), key) )
        return s



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
import unittest
import os
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.bse.bse_absorption import YamboBSEAbsorptionSpectra
test_path = os.path.join(os.path.dirname(__file__),'..','..','data','refs','bse')

class TestYamboBSEAbsorptionSpectra(unittest.TestCase):

    def test_yambobseabsorptionspeectra(self):

        #load databases
        lat  = YamboLatticeDB.from_db_file(os.path.join(test_path,'SAVE','ns.db1'))
        exc  = YamboExcitonDB.from_db_file(lat,folder=os.path.join(test_path,'yambo'))

        #open show analysis of bse absorption spectra
        bsea = YamboBSEAbsorptionSpectra(exc)
        bsea.plot()

        #bsea.write_json()

if __name__ == '__main__':
    unittest.main()



# By FP (2023)

from yambopy import *
from yambopy.lattice import point_matching
from netCDF4 import Dataset

class YamboQPRotate():
    """
    This class expands the ndb.QP computed by yambo in the IBZ to the
    full k-BZ (i.e., Wigner-Seitz cell).

    === Usage and variables ===

    >> yqp  = YamboQPDB.from_db(filename='ndb.QP',folder='db_path')
    >> yexpand = YamboQPRotate(yqp,save_path=db_path,path_output_DBs=db_path.split('/')[0]+'/Expanded',verbose=1)

    :: Input
    - yqp             -> YamboQPDB object (IBZ calculation)
    - save_path       -> [OPTIONAL] Location of ns.db1 database (default is inside SAVE)
    - path_output_DBs -> [OPTIONAL] Print expanded databases at location (default: do not print)
    - verbose         -> [OPTIONAL] If True, prints a list of kpoints for QE calculation of nosym system

    :: Output
    - numpy array with expanded quasiparticle energies
    - [OPTIONAL] ndb.QP database corresponding to the full BZ.
    - [OPTIONAL] data file 'kpoints_bz.dat' containing expanded rlu kpoint coordinates in PW format

    """
    def __init__(self,yqp,save_path="SAVE",db1='ns.db1',path_output_DBs=None,verbose=0):

        # Quantities to store from QPDB
        self.nkibz, self.nbands = yqp.eigenvalues_qp.shape
        self.qp_path            = yqp.qp_path

        self.k_output     = 'kpoints_bz.dat'

        # Read geometry and BZ expanded k-grid
        lattice     = YamboLatticeDB.from_db_file(save_path+"/ns.db1")
        self.map_bz2ibz  = lattice.kpoints_indexes
        self.nkbz        = lattice.nkpoints
        self.alat        = lattice.alat
        self.rlat        = lattice.rlat
        self.iku_kpoints = np.array(lattice.iku_kpoints)

        self.nstates     = self.nkbz*self.nbands

        print("=== Rotating ndb.QP... ===")

        print(" * QP kpts and table... ")
        # Check that IBZ k-grid is complete and matches ns.db1
        if len(yqp.qps['Kpoint'])!=self.nkibz: 
            raise ValueError('\n[ERROR] Expansion currently works if ndb.QP is computed for all IBZ kpoints.')
        k_matches = point_matching(yqp.qps['Kpoint'],lattice.ibz_kpoints)
        if not k_matches.tolist() == [ ik for ik in range(self.nkibz) ]:
             raise TypeError('\n[ERROR] Something wrong with QP kpts. It seems k-indices are not matching with SAVE data.')

        self.qp_table = self.expand_QPtable(yqp.qps['qp_table'])
        print(" * KS energies... ")
        self.Eo = self.expand_QP(yqp.qps['Eo'].real)
        print(" * QP energies and lifetimes... ")
        self.E = self.expand_QP(yqp.qps['E'])
        print(" * Z renorm factors... ")
        self.Z = self.expand_QP(yqp.qps['Z'])

        if path_output_DBs is not None:
            print(" * Saving databases... ")
            self.outpath = path_output_DBs
            self.saveDBS()
        else: print(" [!] Enter value for path_output_DBs to print expanded ndb.QP")

        # Print kpts for PW nosym run
        if verbose: self.print_kpts_PW_format(lattice.iku_kpoints,units='rlu')

        print("===      Done.       ===")

    def expand_QP(self,values_ibz):
        """
        Expand a QP quantity
        """

        values_bz = np.zeros((self.nkbz*self.nbands),dtype=values_ibz.dtype)
        for ib in range(self.nbands):
            for ikbz in range(self.nkbz): 
                bz_index  = ib+ikbz*self.nbands
                ibz_index = ib+self.map_bz2ibz[ikbz]*self.nbands
                values_bz[bz_index]= values_ibz[ibz_index]

        return values_bz

    def expand_QPtable(self,table):
        """
        QP table works like this:
        table[0][i_QP]=i_band1
        table[1][i_QP]=i_band2 (= i_band1 since SE is diagonal)
        table[2][i_QP]=i_kpt
        """
        table_b1  = self.expand_QP(table[0])
        table_b2  = self.expand_QP(table[1])
        table_kpt = np.zeros(self.nstates)

        for ib in range(self.nbands):
            for ikbz in range(self.nkbz):
                bz_index  = ib+ikbz*self.nbands
                table_kpt[bz_index]=ikbz

        table_bz = np.array([table_b1,table_b2,table_kpt])

        return table_bz

    def saveDBS(self):
        """
        Write yambo-compatible ndb.QP databases containing the expanded quasiparticles.

        WARNING:
            There are many quirks and special rules and version-dependent ifs in the yambo IO 
            of ndb.QP.

            The implementation here has to overcome the difficulty of indulging these quirks so that
            the yambopy-created database is not rejected by yambo.

            The present implementation tries to be as general as reasonably possible, and it has been tested,
            but instances may still arise with the appearance of errors and incompatibilities while trying to
            get yambo to read these dbs. Also, in the event of an update in the yambo IO of ndb.QP, these
            writing functions will almost certainly be broken.
        
        ==========================================================================================
        Write ndb.QP

            :: Dimensions ([F]ixed or [V]ariable):
               0 - D_0000000001 = 1   [F]-> QP_DB_kind, HEAD_REVISION, SERIAL_NUMBER, CUTOFF,
                                            G_energies_xc_KIND,G_wavefunctions_xc_KIND,Xp_energies_xc_KIND,
                                            Xp_wavefunctions_xc_KIND,QP_N_DESCRIPTORS,QP_GW_solver,
                                            QP_GW_approximation,QP_PPA_imaginary_Energy,QP_GW_SC_iterations,
                                            QP_dS_dw_steps,QP_dS_dw_step,QP_X_Gs,QP_X_poles,QP_X_xc-Kernel,
                                            QP_X_BZ_energy_Double_Grid,QP_Sc_G_damping,QP_Sc_bands_terminator,
                                            QP_Sx_RL_components,QP_EMPTY_STR_Nr18,QP_GF_energies_kind,
                                            QP_GF_WFs_kind,QP_Xs_energies_kind,QP_Xs_WFs_kind
               1 - D_0000000001 = 3   [F]-> HEAD_VERSION, HEAD_D_LATT, QP_table, QP_kpts
               2 - D_0000000002 = 2   [F]-> SPIN_VARS, TEMPERATURES, QP_X_bands, QP_X_e_h_E_range,
                                            QP_Sc_G_bands,QP_QP_@_state_1_K_range,QP_QP_@_state_1_b_range,
                                            QP_E,QP_Z
               3 - D_0000000100 = 100 [F]-> G_energies_xc_KIND,G_wavefunctions_xc_KIND,Xp_energies_xc_KIND,
                                            Xp_wavefunctions_xc_KIND,QP_GW_solver,QP_GW_approximation,
                                            QP_X_xc-Kernel,QP_EMPTY_STR_Nr18,QP_GF_energies_kind,QP_GF_WFs_kind,
                                            QP_Xs_energies_kind,QP_Xs_WFs_kind
               4 - D_0000000006 = 6   [F]-> PARS
               5 - QP_desc_size = 24  [F]-> QP_DESCRIPTORS_SIZES,QP_DESCRIPTORS_NAMES,QP_DESCRIPTORS_KINDS,
                                            QP_DESCRIPTORS_TERMS
               6 - QP_string_len = 100 [F]-> QP_DESCRIPTORS_NAMES,QP_DESCRIPTORS_KINDS,QP_DESCRIPTORS_TERMS
               7 - D_0000000Nqp = Nqp  [V]-> QP_table,QP_E,QP_Eo,QP_Z
               8 - D_00000000Nk = Nk   [V]-> QP_kpts,QP_E,QP_Eo,QP_Z

            :: Contents of PARS:
                - [F]Nbands, qp%nb
                - [V]Nkpts in IBZ, qp%nk
                - [V]Nqp=Nkpts*Nbands, qp%n_states
                - [F]GWo SC iterations, GWo_iterations
                - [F]GW SC iterations, GW_iterations
                - [F]QP_desc_size, qp%desc%n

        ===========================================================================================
        """
        def netcdftype(var_type):
            """ Distinguish between double and float
            """
            if var.dtype=='float32': return 'f4'
            elif var.dtype=='float64': return 'f8'
            else: raise TypeError('\n[ERROR] Variable type not recognized. It should be either float (float32) or double (float64).\n')

        path = self.outpath
        # If it exists, always remove outdir for safety
        if os.path.isdir(path): shutil.rmtree(path)
        os.mkdir(path)

        # New database
        dbs = Dataset(path+'/ndb.QP',mode='w',format='NETCDF4')

        # Old database
        dbs_ibz = Dataset(self.qp_path+'/ndb.QP')
        ibz_dims = dbs_ibz.dimensions.values()
        ibz_vars = dbs_ibz.variables.values()

        # check if we are in an unfortunate cases due to how yambo writes the databases
        sizes     = [dim.size for dim in ibz_dims]
        hard_dims = sizes[:-2] # fixed dimension sizes
        ibz_ = self.nkibz   in hard_dims #k_ibz number coincides with one of the fixed dimensions
        iqp_ = self.nkibz*self.nbands in hard_dims #qp ibz number coincides with one of the fixed dimensions
        bz_  = self.nkbz    in hard_dims #k_bz number coincides with one of the fixed dimensions
        qp_  = self.nstates in hard_dims #qp number coincides with one of the fixed dimensions

        sizes     = [dim.size for dim in ibz_dims]
        n_dims    = len(sizes)
        ind_k_ibz_dim = sizes.index(self.nkibz)
        # Nk-related dimension is in last place, Nqp-related one in 7th place
        error_conditions=np.array([n_dims==9 and ind_k_ibz_dim!=8, n_dims<9 and ibz_==False,n_dims<9 and iqp_==False])
        if error_conditions.any(): raise ValueError("[ERROR] something wrong with QP dbs dimensions.")        

        # Create dimensions
        iaux=0
        for dim in ibz_dims:
            # new k_bz dimension
            if iaux==6 and qp_==False: dbs.createDimension('D_%010d'%self.nstates,self.nstates)
            elif iaux==6 and qp_==True: continue
            if iaux==7 and bz_==False: dbs.createDimension('D_%010d'%self.nkbz,self.nkbz)
            elif iaux==7 and bz_==True: continue
            # fixed dimensions
            else: dbs.createDimension(dim.name,dim.size)
            iaux+=1

        # New variables
        pars = dbs_ibz.variables['PARS'][:]
        pars[1] = self.nkbz
        pars[2] = self.nstates 

        QP_E = np.stack((self.E.real, self.E.imag),axis=1)
        QP_Z = np.stack((self.Z.real, self.Z.imag),axis=1)

        # Create variables
        for var in ibz_vars:
            if var.name=='QP_table': dbs.createVariable('QP_table',netcdftype(var.dtype), ('D_%.10d'%3, 'D_%.10d'%self.nstates))
            elif var.name=='QP_kpts': dbs.createVariable('QP_kpts',netcdftype(var.dtype), ('D_%.10d'%3, 'D_%.10d'%self.nkbz))
            elif var.name=='QP_E': dbs.createVariable('QP_E',netcdftype(var.dtype), ('D_%.10d'%self.nstates, 'D_%.10d'%2))
            elif var.name=='QP_Eo': dbs.createVariable('QP_Eo',netcdftype(var.dtype), ('D_%.10d'%self.nstates))
            elif var.name=='QP_Z': dbs.createVariable('QP_Z',netcdftype(var.dtype), ('D_%.10d'%self.nstates, 'D_%.10d'%2))
            else: dbs.createVariable(var.name, var.dtype, var.dimensions)

        # Store values in new DB, including new qpt coords in iku
        for var in ibz_vars:
            if   var.name=='PARS':     dbs[var.name][:] = pars
            elif var.name=='QP_table': dbs[var.name][:] = self.qp_table
            elif var.name=='QP_kpts':  dbs[var.name][:] = self.iku_kpoints.T
            elif var.name=='QP_E':     dbs[var.name][:] = QP_E
            elif var.name=='QP_Eo':    dbs[var.name][:] = self.Eo
            elif var.name=='QP_Z':     dbs[var.name][:] = QP_Z
            elif var.name=='QP_QP_@_state_1_K_range': dbs[var.name][:] = [ 1, self.nkbz ]
            else:                      dbs[var.name][:] = dbs_ibz[var.name][:]

        dbs.close()
        dbs_ibz.close()

    def print_kpts_PW_format(self,kpoints_iku,units='rlu'):
        """
        Print expanded kpoints in the Wigner-Seitz cell for PW
        nosym=.true., noinv=.true. calculation

        - kpoints_iku: iku kpoints in the BZ read from ns.db1
        - units: either rlu or cc
        """
        def format_array(value):
            if value >= 0: return ' '+format(value,'.9f')
            else:          return format(value,'.9f')

        print(" * Printing PW-format kpoints file.")

        kpoints = np.array([ k/self.alat for k in kpoints_iku ])
        if units=='rlu': points = car_red(kpoints,self.rlat)
        if units=='cc':  points = kpoints*self.alat[0]

        kpts2prnt = np.empty((len(points),4),dtype=object)
        for i in range(len(points)):
            for j in range(4):
                if j<3:
                    val = points[i,j]
                    if val>=0: kpts2prnt[i,j]=' '+"{:0.9f}".format(val)
                    else:      kpts2prnt[i,j]="{:0.9f}".format(val)
                if j==3:       kpts2prnt[i,j]='1'

        np.savetxt(self.k_output, kpts2prnt, fmt='%s %s %s %s')

    def __str__(self):

        lines = []; app=lines.append
        app(marquee(self.__class__.__name__))

        app('nkpoints (ibz):  %d'%self.nkibz)
        app('nkpoints (bz):   %d'%self.nkbz)
        app('nbands:          %d'%self.nbands)
        app('nqpstates (ibz): %d'%(self.nkibz*self.nbands))
        app('nqpstates (bz):  %d'%self.nstates)

        return "\n".join(lines)



# This file is part of yambopy
# Author: FP
from yambopy import *
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from yambopy.units import ha2ev, ev2cm1, I
from yambopy.plot.plotting import add_fig_kwargs,BZ_hexagon,shifted_grids_2D

@add_fig_kwargs
def plot_BZ_2D(self,data,bzgrid_cc=None,plt_show=False,plt_cbar=False,**kwargs):
    """
    2D scatterplot in the BZ of the quantity A_{k}(i1,i2,i3,...).

    Any real quantity which is a function of only the k-grid may be supplied.
    The additional indices are user-specified.

    Example:
    - Exciton coefficients A_{k}(iq,ilambda,iv,ic) or A_{q}(ik,ilambda,iv,ic)
    - Electron-phonon matrix elements g_{k}(iq,imu,ib1,ib2) or g_{q}(ik,imu,ib1,ib2)
    - In general any k- or q-dependent quantity

    Inputs:
    - latticeDB object (lattice)
    - data do visualise    
    - BZ-grid in Cartesian coordinates (if not provided, uses lat.car_kpoints)

    Options:
    - if plt_show plot is shown
    - if plt_cbar colorbar is shown
    - kwargs example: marker='H', s=300, cmap='viridis', etc.

    Returns:
    - fig and ax objects

    NB: so far works for hexagonal systems. Can be improved to plot BZ planes at constant k_z for 3D systems and non-hexagonal cells.
    """
    rlat = lattice.rlat
    if bzgrid_cc is not None: 
        kpts = bz_grid_cc
    else:
        print("Using yambopy kpoints for plot") 
        kpts = self.car_kpoints

    # Input check
    if len(data)!=len(kpts):
        raise ValueError('Something wrong in data dimensions (%d data vs %d kpts)'%(len(data),len(kpts)))

    # Global plot stuff
    fig, ax = plt.subplots(1, 1)
    ax.add_patch(BZ_hexagon(rlat))

    if plt_cbar:
        if 'cmap' in kwargs.keys(): color_map = plt.get_cmap(kwargs['cmap'])
        else:                       color_map = plt.get_cmap('viridis')
    lim = 1.05*np.linalg.norm(rlat[0])
    ax.set_xlim(-lim,lim)
    ax.set_ylim(-lim,lim)

    # Reproduce plot also in adjacent BZs
    BZs = shifted_grids_2D(kpts,rlat)
    for kpts_s in BZs: plot=ax.scatter(kpts_s[:,0],kpts_s[:,1],c=data,**kwargs)

    if plt_cbar: fig.colorbar(plot)

    plt.gca().set_aspect('equal')

    if plt_show: plt.show()
    else: print("Plot ready.\nYou can customise adding savefig, title, labels, text, show, etc...")

    return fig, ax

class ExpandMatrixElement():
    """
    This is a generic symmetry expansion script.

    Input: 
      - System symmetries from YamboLatticeDB
      - Array of matrix elements on k/q-grids to be expanded from IBZ to BZ
      - Instruction to expand over k or over q

    Output:
      - Array of expanded matrix elements in the specified grid

    Example:

      :: O_nmR{k}^S{q} =  <n R{k}|O|mR{k}-S{q}>

      :: Expansion over q is

      :: O_nmR{k}^S{q} =  O_nm S^-1R{k}^q     (no TR)
      :: O_nmR{k}^S{q} = [O_nm S^-1R{k}^q]^*  (TR)

      :: Expansion over k is

      :: O_nmR{k}^S{q} =  O_nm k^R^-1S{q}     (no TR)
      :: O_nmR{k}^S{q} = [O_nm k^R^-1S{q}]^*  (TR)

    """
    #def __init__(mats_ibz,syms,space='q',TR=False):
        
